#ifdef HAVE_BUFFER
#error Can not have more than one BUFFER module.
#endif

#define HAVE_BUFFER
#include "buffer/interface.h"
#include "emulation/interface.h"
#include DISPLAY
#include <assert.h>
#include <stdio.h>
#include "global.h"


extern buf_cell_t bufdata[80 * 25];
extern buf_cell_t *rwptr;
extern buf_state_t bufstate;
extern void *bufemu;

/* invalidate display */
void buf_update(const void *b, const int x1, const int y1, const int x2, const int y2);

/* hide cursor */
static inline void buf_cursor_hide(const void *b)
{
	display_cursor_hide();
}

/* set internal r/w pointer in (visible) buffer */
static inline void buf_set_rw(const void *b, const int x, const int y)
{
	assert(0<=x && x<80 && 0<=y && y<25);
	rwptr = bufdata + y * 80 + x;
}

static inline void buf_set_r(const void *b, const int x, const int y)
{
	buf_set_rw(b, x, y);
}


/* set cursor position. sets r/w pointer to the same position */
static inline void buf_set_cursor(const void *b, const int x, const int y)
{
	assert(0<=x && x<80 && 0<=y && y<25);

	buf_cursor_hide(b);

	bufstate.cursor_x = x;
	bufstate.cursor_y = y;
	buf_set_rw(b, x, y);

	buf_update(b, x, x, y, y);
}


/* display->buffer: creates new buffer, returns handle.
 * buffer will create emu and io */
static inline void *buf_create()
{
	buf_cell_t *p;
	int i;

	p = bufdata;
	for(i=0; i<80*25; i++) {
		p->fg = 7;
		p->bg = 0;
		p->c = 0;
		p->attr = 0;
		p++;
	}

	bufstate.w = 80;
	bufstate.h = 25;
	bufstate.fg = 7;
	bufstate.bg = 0;
	bufstate.attr = 0;
	bufstate.cursor_x = 0;
	bufstate.cursor_y = 0;
	bufstate.cursor_type = buf_cursor_block;
	bufstate.scrv = 0;
	buf_set_rw((void *) 1, 0, 0);

	bufemu = emu_create((void *)1, NOECHO);

	return (void *)1;
}

/* display->buffer: set screen dimensions (visible buffer size) */
void buf_set_screen_dim(const int w, const int h);

/* move rw to the cursor position */
static inline void buf_rw_from_cursor(const void *b)
{
	rwptr = bufdata + bufstate.cursor_y * 80 + bufstate.cursor_x;
}

/* move r/w left by one position */
static inline void buf_rw_left()
{
	assert(rwptr > bufdata);
	rwptr--;
}

/* move r/w right by one position */
static inline void buf_rw_right()
{
	assert(rwptr < bufdata + 2000*sizeof(buf_cell_t));
	rwptr++;
}

/* reads next character cell from (visible) buffer, advances r/w pointer
 * returns buf_cell_t
 */
static inline buf_cell_t buf_read_cell(const void *b)
{
	return *(rwptr++);
}

/* writes character (buf_char_t) into buffer, advances r/w pointer
 * uses current colors/attributes
 */
static inline void buf_write_char(const void *b, buf_char_t ch)
{
	rwptr->c = ch;
	rwptr->color = bufstate.color;
	rwptr++;
}

/* returns pointer to state of current buffer
 * buf_state_t *
 * don't write */
static inline buf_state_t * buf_get_state(const void *b)
{
	return &bufstate;
}

/* move cursor and r/w left by one position */
static inline void buf_cursor_left()
{
	bufstate.cursor_x--;
	buf_rw_left();
}

/* move cursor and r/w right by one position */
static inline void buf_cursor_right()
{
	bufstate.cursor_x++;
	buf_rw_right();
}

/* move cursor and r/w up by one position */
static inline void buf_cursor_up()
{
	bufstate.cursor_y--;
	buf_rw_up();
}

/* move cursor and r/w down by one position */
static inline void buf_cursor_down()
{
	bufstate.cursor_y++;
	buf_rw_down();
}

/* set foreground color (char) */
static inline void buf_set_fg(const void *b, unsigned char fg)
{
	bufstate.color.fg = fg;
}

/* set background color (char) */
static inline void buf_set_bg(const void *b, unsigned char bg)
{
	bufstate.color.bg = bg;
}

/* set attrs (char) */
static inline void buf_set_attr(const void *b, buf_attr_t attr)
{
	bufstate.color.attr = attr;
}

/* set color */
static inline void buf_set_color(const void *b, buf_color_t color)
{
	bufstate.color = color;
}

/* set color */
static inline void buf_reset_color(const void *b, buf_color_t color)
{
}

/* process multiple key press events (sends to emu) */
static inline void buf_dsp_send(const void *b, char *s)
{
	char *p;
	for(p = s; *p; p++)
		emu_dsp_char(bufemu, *p);
}

/* process key press event (sends to emu) */
static inline void buf_dsp_char(const void *b, char c)
{
	emu_dsp_char(bufemu, c);
}

/* process ctrl key press event (sends to emu) */
static inline void buf_dsp_ctrl(const void *b, ctrl_t c)
{
	emu_dsp_ctrl(bufemu, c);
}

/* scroll the buffer. bottom line will be filled with 0, default fg/bg/attr */
void buf_scroll(const void *b);

/* adjust buffer scroll */
static inline void buf_adjust_scroll(buf_t *b, int amount)
{
}

/* switch buffer relatively */
static inline void * buf_switch_rel(const int o)
{
	return (void *) 1;
}

static inline void buf_dead(void *b)
{
	exit(0);
}

/* erase part of buffer */
void buf_erase(void *b, const int x1, const int y1, const int x2, const int y2);

