#ifndef display_interface_h
#define display_interface_h

/* type of watch callback */
typedef void (* watched_callback_t) (int, int, void *);

/* structure for watched fds */
typedef struct {
	int fd;
	int wtype;
	watched_callback_t callback;
	void *data;
} watched;

/* size of fd watch array */
#ifndef MAXWATCHED
#define MAXWATCHED 32
#endif

extern watched watches[MAXWATCHED];
extern int numwatched;

/* display pointer to buffer in focus */
extern void * dsp_bf;

/**
 * Start DISPLAY
 * Doesn't return until the DISPLAY exits.
 */
void display_start();

/**
 * register fd for watching
 * wtype: 1: watch for read
 *        2: watch for write
 *        4: watch for exceptions
 */
void display_watch(int fd, int wtype, watched_callback_t callback, void *data);

/**
 * redraw part of display
 */
void display_update(const int x1, const int y1, const int x2, const int y2);

/**
 * unregister fd
 */
void display_unwatch(int fd);

/**
 * set display title
 */
void display_set_title(char *title);

/**
 * scroll display one line up
 */
void display_scroll();

/**
 * shift an area of the display
 */
void display_shift(const int x1, const int y1, const int x2, const int y2, const int xd, const int yd);

/**
 * suicide
 */
void display_exit();

/**
 * ring the bell
 */
void display_bell();

/**
 * hide cursor
 */
void display_cursor_hide();

/**
 * set active buffer
 */
void display_set_buffer(void *b);

/**
 * reports the actual size of the terminal in characters
 */
void display_get_size(int *rows, int *cols);

/**
 * Cursor settings
 */
typedef enum dsp_curs_types_e {
	CURS_BLOCK,
	CURS_COLUMN,
	CURS_VGA,
	CURS_EMPTYBLOCK,
} dsp_curs_types_t;


#ifndef HAVE_DISPLAY
#error No display is selected.
#endif

#endif /*display_interface_h*/
