#include <stdlib.h>
#include <string.h>

#include "linux.h"
#include "internal.h"

#include "actions.c"
#include "transmit.c"
#include "process.c"

/*

linux console emulation

*/

#include <asm/ioctls.h>
#include <stropts.h>
#include <termios.h>
void emu_setup_fd(int fd)
{
	struct termios term;

	memset((char *)&term,0,sizeof(term));
	term.c_iflag = BRKINT | IGNPAR | ICRNL | IXON;
	term.c_oflag = OPOST | ONLCR;
	term.c_cflag = B9600 | CREAD | CS8;
	term.c_lflag = ISIG | IEXTEN | ICANON | ECHO | ECHOE | ECHOK;

	term.c_cc[VERASE]   = 0177;  /* DEL */
	term.c_cc[VDISCARD] = 017;   /* ^O */
	term.c_cc[VSUSP]    = 032;   /* ^Z */
	term.c_cc[VINTR]    = 003;   /* ^C */
	term.c_cc[VQUIT]    = 034;   /* ^\ */
	term.c_cc[VEOF]     = 004;   /* ^D */
	term.c_cc[VKILL]    = 025;   /* ^U */
	term.c_cc[VSTART]   = 021;   /* ^Q */
	term.c_cc[VSTOP]    = 023;   /* ^S */
	tcsetattr(fd,TCSANOW,&term);
	if (ioctl(fd,TIOCCONS,fd) != 0)
		log_printf(1, "Emulation: Could not set console\n");
}

void *emu_create(void *buf)
{
	emu_t *e = malloc(sizeof(emu_t));
	if (e == NULL) {
		return NULL;
	}
	e->buf = buf;
	e->bs = buf_get_state(buf);
	e->io = io_create(e);
	emu_reset_terminal(e);
	emu_clear_screen(e);
	return (void *)e;
}

void emu_destroy(void *emu)
{
	emu_t *e = emu;
	io_destroy(e->io);
	free(e);
}

static void emu_reset_terminal(emu_t *e)
{
	buf_reset_color(e->buf);
	
	e->top          = 0;
	e->bottom       = e->bs->h - 1;
	e->es           = ES_NORMAL;
	e->question     = 0;
	e->need_wrap    = 0;
	e->crlf		= 0;
	e->insert       = 0;
//	e->translate    = set_translate(LAT1_MAP, vc);
//	e->G0_charset   = LAT1_MAP;
//	e->G1_charset   = GRAF_MAP;
	e->charset      = 0;
	e->saved_charset= 0;
//	e->utf          = 0;
//	e->utf_count    = 0;
	e->show_control = 0;
	e->toggle_meta  = 0;
	e->decckm       = 0;
	e->deccolm      = 0;
	e->decscnm      = 0;
	e->decom        = 0;
	e->decawm       = 1;
	e->decarm       = 1;
	e->mouse_x10    = 0;
	e->mouse_x11    = 0;
	e->deccm        = 1;

	e->tab_stop[0]  = 0x01010100;
	e->tab_stop[1]  =
	e->tab_stop[2]  =
	e->tab_stop[3]  =
	e->tab_stop[4]  = 0x01010101;
//	e->bell_pitch = DEFAULT_BELL_PITCH;
//	e->bell_duration = DEFAULT_BELL_DURATION;
	emu_gotoxy(e, 0, 0);
	emu_save_cur(e);
}

