/*

process recieved characters (main logic, calls functions of actions.c)

reference: man consol_codes, ECMA-48 (== ISO/IES 6429)

*/

static inline void emu_csi_J(emu_t *e, int param)
{
	switch (param) {
		case 0:	/* erase from cursor to end of display */
			emu_clear_chars(e, e->bs->cursor_x, e->bs->w - 1);
			if (e->bs->cursor_y < e->bs->h - 1) {
				emu_clear_lines(e, e->bs->cursor_y + 1, e->bs->h - 1);
			}
			break;
		case 1:	/* erase from start to cursor */
			emu_clear_chars(e, 0, e->bs->cursor_x);
			if (e->bs->cursor_y > 0) {
				emu_clear_lines(e, 0, e->bs->cursor_y - 1);
			}
			break;
		case 2: /* erase whole display */
			emu_clear_screen(e);
			break;
		default:
			return;
	}
	e->need_wrap = 0;
}

static inline void emu_csi_K(emu_t *e, int param)
{
	switch (param) {
		case 0:	/* erase from cursor to end of line */
			emu_clear_chars(e, e->bs->cursor_x, e->bs->w - 1);
			break;
		case 1:	/* erase from start of line to cursor */
			emu_clear_chars(e, 0, e->bs->cursor_x);
			break;
		case 2: /* erase whole line */
			emu_clear_chars(e, 0, e->bs->w - 1);
			break;
		default:
			return;
	}
	e->need_wrap = 0;
}

/* SGR - set graphics rendition */
static inline void emu_csi_m(emu_t *e)
{
	int i;
	
	/* TODO: use buf_set_fg, buf_set_bg, buf_set_color, buf_set_attr.. */
	for (i = 0; i <= e->nparam; i++) {
		switch (e->param[i]) {
			case 0:
				e->bs->color = e->bs->default_color;
				break;
			case 1:
				e->bs->color.attr.bold = 1;
				e->bs->color.attr.halfbright = 0;
				break;
			case 2:
				e->bs->color.attr.bold = 0;
				e->bs->color.attr.halfbright = 1;
				break;
			case 4:
				e->bs->color.attr.underline = 1;
				break;
			case 5:
				e->bs->color.attr.blink = 1;
				break;
			case 7:
				e->bs->color.attr.reverse = 1;
				break;
			case 10:
				/* ANSI X3.64-1979 (SCO-ish?) */
/*
				e->vc_translate = set_translate(e->vc_charset == 0
						? e->vc_G0_charset
						: e->vc_G1_charset, vc);
*/
				e->show_control = 0;
				e->toggle_meta = 0;
				break;
			case 11:
				/* ANSI X3.64-1979 (SCO-ish?) */
//				e->vc_translate = set_translate(IBMPC_MAP, vc);
				e->show_control = 1;
				e->toggle_meta = 0;
				break;
			case 12:
				/* ANSI X3.64-1979 (SCO-ish?) */
//				e->vc_translate = set_translate(IBMPC_MAP, vc);
				e->show_control = 1;
				e->toggle_meta = 1;
				break;
			case 21: /* incompatible with ECMA-48 */
			case 22:
				e->bs->color.attr.bold = 0;
				e->bs->color.attr.halfbright = 0;
				break;
			case 24:
				e->bs->color.attr.underline = 0;
				break;
			case 25:
				e->bs->color.attr.blink = 0;
				break;
			case 27:
				e->bs->color.attr.reverse = 0;
				break;
			case 38:
				/* ANSI X3.64-1979 (SCO-ish?) */
				e->bs->color.fg = e->bs->default_color.fg;
				e->bs->color.attr.underline = 1;
				break;
			case 39:
				/* ANSI X3.64-1979 (SCO-ish?) */
				e->bs->color.fg = e->bs->default_color.fg;
				e->bs->color.attr.underline = 0;
				break;
			case 49:
				e->bs->color.bg = e->bs->default_color.bg;
				break;
			default:
				if (e->param[i] >= 30 && e->param[i] <= 37)
					e->bs->color.fg = e->param[i] - 30;
				else if (e->param[i] >= 40 && e->param[i] <= 47)
					e->bs->color.bg = e->param[i] - 40;
				break;
		}
	}
}

static inline void emu_set_mode(emu_t *e, int on_off)
{
	int i;
	for (i = 0; i <= e->nparam; i++) {
		if (e->question) {
			/* DEC Private Modes */
			switch(e->param[i]) {
			case 1:	
				NI("DECCKM (cursor sends an additional 0)");
				e->decckm = on_off;
				break;
			case 3:
				NI("DECCOLM 80/132 columns");
				e->deccolm = on_off;
				break;
			case 5:
				NI("DECSCNM invert screen");
				if (e->decscnm != on_off) {
					e->decscnm = on_off;
					/*
					emu_invert_screen(e, 0, e->vc_screenbuf_size, 0);
					emu_update_attr(e);
					*/
				}
				break;
			case 6:
				D("DECOM orig relative");
				e->decom = on_off;
				emu_gotoxay(e, 0, 0);
				break;
			case 7:
				NI("DECAWM autowrap");
				e->decawm = on_off;
				break;
			case 8:
				NI("DECARM autorepeat");
				e->decarm = on_off;
				break;
			case 9:
				NI("X10 mouse reporting");
				e->mouse_x10 = on_off;
				break;
			case 25:
				NI("DECCM show mouse cursor");
				e->deccm = on_off;
				break;
			case 1000:
				NI("X11 mouse reporting");
				e->mouse_x11 = on_off;
				break;
			}
		} else {
			switch(e->param[i]) {
			case 3: /* Monitor (display ctrls) */
				e->show_control = on_off;
				break;
			case 4: /* Insert Mode on/off */
				e->insert = on_off;
				break;
			case 20: /* Lf, Enter == CrLf/Lf */
				e->crlf = on_off;
			}
		}
	}
}

static inline void emu_setterm_command(emu_t *e)
{
	switch(e->param[0]) {
		case 1:	/* set color for underline mode */
			D("underline");
			if (e->param[1] < 16) {
				e->bs->underline_fg = e->param[1];
			}
			break;
		case 2:	/* set color for half intensity mode */
			D("half intensity");
			if (e->param[1] < 16) {
				e->bs->halfbright_fg = e->param[1];
			}
			break;
		case 8:	/* store colors as defaults */
			D("store color as default");
			e->bs->default_color = e->bs->color;
			// TODO: attrs ? buf update ?
			break;
		case 9:	/* set blanking interval */
			NI("blank interval");
			//blankinterval = ((e->vc_par[1] < 60) ? e->vc_par[1] : 60) * 60 * HZ;
			//poke_blanked_console();
			break;
		case 10: /* set bell frequency in Hz */
			NI("bell freq");
			//if (e->vc_npar >= 1)
			//	e->vc_bell_pitch = e->vc_par[1];
			//else
			//	e->vc_bell_pitch = DEFAULT_BELL_PITCH;
			break;
		case 11: /* set bell duration in msec */
			NI("bell duration");
			//if (e->vc_npar >= 1)
			//	e->vc_bell_duration = (e->vc_par[1] < 2000) ?
			//		e->vc_par[1] * HZ / 1000 : 0;
			//else
			//	e->vc_bell_duration = DEFAULT_BELL_DURATION;
			break;
		case 12: /* bring specified console to the front */
			NI("change console");
			//if (e->vc_par[1] >= 1 && vc_cons_allocated(e->vc_par[1] - 1))
			//	set_console(e->vc_par[1] - 1);
			break;
		case 13: /* unblank the screen */
			NI("unblank screen");
			//poke_blanked_console();
			break;
		case 14: /* set vesa powerdown interval */
			NI("powerdown interval");
			//vesa_off_interval = ((e->vc_par[1] < 60) ? e->vc_par[1] : 60) * 60 * HZ;
			break;
		case 15: /* activate the previous console */
			NI("last console");
			//set_console(last_console);
			break;
	}
}

static inline void emu_process_params(emu_t *e, emu_char_t c)
{
//	log_printf(2,"ESC [%s * %c  *=%d;%d;%d;... nparams=%d\n", e->question?" ?":"", c, e->param[0], e->param[1], e->param[2], e->nparam + 1);
	e->es = ES_NORMAL;
	switch (c) {
	case 'h':
		emu_set_mode(e, 1);
		return;
	case 'l':
		emu_set_mode(e, 0);
		return;
	case 'c':
		if (e->question) {
			NI("set cursor type");
			// TODO: cursor type
			//cursor_type = e->param[0] | (e->param[1] << 8) | (e->param[2] << 16);
			switch(e->param[0]) {
			case 0:
				e->bs->cursor_type = buf_cursor_block;
				return;
			case 1:
				e->bs->cursor_type = buf_cursor_none;
				return;
			case 2: //underline
			case 3: //lower third
			case 4: //lower half
			case 5: //two third
			case 6: // block
				e->bs->cursor_type = buf_cursor_block;
				return;
			}
			return;
		}
		break;
	case 'm':
		if (e->question) {
			NI("clear selection");
/*
			emu_clear_selection();
			if (e->vc_par[0])
				e->vc_complement_mask = e->vc_par[0] << 8 | e->vc_par[1];
			else
				e->vc_complement_mask = e->vc_s_complement_mask;
*/
			return;
		}
		break;
	case 'n':
		if (!e->question) {
			if (e->param[0] == 5)
				emu_status_report(e);
			else if (e->param[0] == 6)
				emu_cursor_report(e);
		}
		return;
	}
	
	if (e->question) {
		e->question = 0;
		return;
	}
	switch (c) {
	// TODO: optimize (left, right, up, down instead of goto)
	case 'G': case '`':
		if (e->param[0])
			e->param[0]--;
		emu_gotoxy(e, e->param[0], e->bs->cursor_y);
		return;
	case 'A':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, e->bs->cursor_x, e->bs->cursor_y - e->param[0]);
		return;
	case 'B': case 'e':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, e->bs->cursor_x, e->bs->cursor_y + e->param[0]);
		return;
	case 'C': case 'a':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, e->bs->cursor_x + e->param[0], e->bs->cursor_y);
		return;
	case 'D':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, e->bs->cursor_x - e->param[0], e->bs->cursor_y);
		return;
	case 'E':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, 0, e->bs->cursor_y + e->param[0]);
		return;
	case 'F':
		if (!e->param[0])
			e->param[0]++;
		emu_gotoxy(e, 0, e->bs->cursor_y - e->param[0]);
		return;
	case 'd':
		if (e->param[0])
			e->param[0]--;
		emu_gotoxay(e, e->bs->cursor_x ,e->param[0]);
		return;
	case 'H': case 'f':
		if (e->param[0])
			e->param[0]--;
		if (e->param[1])
			e->param[1]--;
		emu_gotoxay(e, e->param[1], e->param[0]);
		return;
	case 'J':
		emu_csi_J(e, e->param[0]);
		return;
	case 'K':
		emu_csi_K(e, e->param[0]);
		return;
	case 'L':
		if (!e->param[0])
			e->param[0]++;
		emu_scrolldown(e, e->bs->cursor_y, e->bottom, e->param[0]);
		return;
	case 'M':
		if (!e->param[0])
			e->param[0]++;
		emu_scrollup(e, e->bs->cursor_y, e->bottom, e->param[0]);
		return;
	case 'P':
		if (!e->param[0])
			e->param[0]++;
		emu_delete_chars(e, e->param[0]);
		return;
	case 'c':
		if (!e->param[0])
			emu_respond_id(e);
		return;
	case 'g':
		if (!e->param[0])
			e->tab_stop[e->bs->cursor_x >> 5] &= ~(1 << (e->bs->cursor_x & 31));
		else if (e->param[0] == 3) {
			e->tab_stop[0] =
			e->tab_stop[1] =
			e->tab_stop[2] =
			e->tab_stop[3] =
			e->tab_stop[4] = 0;
		}
		return;
	case 'm':
		emu_csi_m(e);
		return;
	case 'q':
		NI("DECLL set keyboard leds");
		return;
	case 'r':
		if (!e->param[0])
			e->param[0]++;
		if (!e->param[1])
			e->param[1] = e->bs->h;
		if (e->param[0] < e->param[1] && e->param[1] <= e->bs->h) {
			e->top = e->param[0] - 1;
			e->bottom = e->param[1] - 1;
			emu_gotoxay(e, 0, 0);
		}
		return;
	case 's':
		D("save cursor");
		emu_save_cur(e);
		return;
	case 'u':
		emu_restore_cur(e);
		return;
	case 'X': /* clear next n chars */
		if (!e->param[0])
			e->param[0]++;
		e->param[0] += e->bs->cursor_x - 1;
		// TODO: move this check to clear_chars ?
		if (e->param[0] >= e->bs->w) {
			e->param[0] = e->bs->w - 1;
		}
		emu_clear_chars(e, e->bs->cursor_x, e->param[0]);
		e->need_wrap = 0;
		return;
	case '@':
		if (!e->param[0])
			e->param[0]++;
		emu_insert_chars(e, e->param[0]);
		return;
	case ']': /* setterm functions */
		emu_setterm_command(e);
		return;
	}
}

static inline void emu_process(emu_t *e, emu_char_t c)
{
	int i;
	switch ((emu_control_t)c) {
	case NUL:
		D("NUL");
		return;
	case BEL:
		D("BEL");
		display_bell();
		return;
	case BS:
		emu_bs(e);
		return;
	case HT:
		D("HT");
		i = e->bs->cursor_x;
		while (i < e->bs->w - 1) {
			i++;
			if (e->tab_stop[i >> 5] & (1 << (i & 31)))
				break;
		}
		buf_set_cursor(e->buf, i, e->bs->cursor_y);
		return;
	case LF: case VT: case FF:
		emu_lf(e);
		if (!e->crlf)
			return;
	case CR:
		emu_cr(e);
		return;
	case SO:
		NI("SO");
//		e->vc_charset = 1;
//		e->vc_translate = set_translate(e->vc_G1_charset, vc);
		e->show_control = 1;
		return;
	case SI:
		NI("SI");
//		e->vc_charset = 0;
//		e->vc_translate = set_translate(e->vc_G0_charset, vc);
		e->show_control = 0;
		return;
	case CAN: case SUB:
		D("CAN/SUB");
		e->es = ES_NORMAL;
		return;
	case ESC:
		e->es = ES_ESC;
		return;
	case DEL:
		D("DEL");
		return;
	case CSI:
		D("CSI");
		e->es = ES_SQUARE;
		return;
	}
	switch(e->es) {
	case ES_ESC:
		e->es = ES_NORMAL;
		switch (c) {
		case '[':
			e->es = ES_SQUARE;
			return;
		case ']':
			e->es = ES_NONSTD;
			return;
		case '%':
			e->es = ES_PERCENT;
			return;
		case 'E':
			emu_cr(e);
			emu_lf(e);
			return;
		case 'M':
			emu_ri(e);
			return;
		case 'D':
			emu_lf(e);
			return;
		case 'H':
			e->tab_stop[e->bs->cursor_x >> 5] |= (1 << (e->bs->cursor_x & 31));
			return;
		case 'Z':
			emu_respond_id(e);
			return;
		case '7':
			emu_save_cur(e);
			return;
		case '8':
			emu_restore_cur(e);
			return;
		case '(':
			e->es = ES_SET_G0;
			return;
		case ')':
			e->es = ES_SET_G1;
			return;
		case '#':
			e->es = ES_HASH;
			return;
		case 'c':
			emu_reset_terminal(e);
			emu_clear_screen(e);
			return;
		case '>':  /* Numeric keypad */
			NI("clear keypad applic mode");
			return;
		case '=':  /* Appl. keypad */
			NI("set keypad applic mode");
			return;
		}
		return;
	case ES_NONSTD:
		if (c=='P') {   /* palette escape sequence */
			for (i = 0; i < NPARAM; i++)
				e->param[i] = 0;
			e->nparam = 0;
			e->es = ES_PALETTE;
			return;
		} else if (c=='R') {   /* reset palette */
			NI("reset palette");
			e->es = ES_NORMAL;
		} else
			e->es = ES_NORMAL;
		return;
	case ES_PALETTE:
		if ( (c>='0' && c<='9') || (c>='A' && c<='F') || (c>='a' && c<='f') ) {
			e->param[e->nparam++] = (c > '9' ? (c & 0xDF) - 'A' + 10 : c - '0');
			if (e->nparam == 7) {
				NI("set palette");
/*				i = e->param[0] * 3;
				j = 1;
				e->palette[i] = 16 * e->param[j++];
				e->palette[i++] += e->param[j++];
				e->palette[i] = 16 * e->param[j++];
				e->palette[i++] += e->param[j++];
				e->palette[i] = 16 * e->param[j++];
				e->palette[i] += e->param[j];
				emu_set_palette(e);
*/
				e->es = ES_NORMAL;
			}
		} else
			e->es = ES_NORMAL;
		return;
	case ES_SQUARE:
		for (i = 0; i < NPARAM; i++)
			e->param[i] = 0;
		e->nparam = 0;
		e->es = ES_GETPARAMS;
		if (c == '[') { /* Function key */
			e->es = ES_FUNCKEY;
			return;
		}
		e->question = (c == '?');
		if (e->question)
			return;
	case ES_GETPARAMS:
		if (c == ';' && e->nparam < NPARAM - 1) {
			e->nparam++;
		} else if (c >= '0' && c <= '9') {
			e->param[e->nparam] *= 10;
			e->param[e->nparam] += c - '0';
		} else
			emu_process_params(e, c);
		return;
	case ES_PERCENT:
		e->es = ES_NORMAL;
		switch (c) {
		case '@':  /* defined in ISO 2022 */
			NI("unset utf8");
//			e->vc_utf = 0;
			return;
		case 'G':  /* prelim official escape code */
		case '8':  /* retained for compatibility */
			NI("set utf8");
//			e->vc_utf = 1;
			return;
		}
		return;
	case ES_FUNCKEY:
		e->es = ES_NORMAL;
		return;
	case ES_HASH:
		e->es = ES_NORMAL;
		if (c == '8') {
			NI("DEC screen alignment test. kludge :-)");
			/* DEC screen alignment test. kludge :-) */
/*
			e->vc_video_erase_char =
				(e->vc_video_erase_char & 0xff00) | 'E';
			emu_csi_J(e, 2);
			e->vc_video_erase_char =
				(e->vc_video_erase_char & 0xff00) | ' ';
			emu_do_update_region(e, e->vc_origin, e->vc_screenbuf_size / 2);
*/
		}
		return;
	case ES_SET_G0:
/*
		if (c == '0')
			e->vc_G0_charset = GRAF_MAP;
		else if (c == 'B')
			e->vc_G0_charset = LAT1_MAP;
		else if (c == 'U')
			e->vc_G0_charset = IBMPC_MAP;
		else if (c == 'K')
			e->vc_G0_charset = USER_MAP;
		if (e->vc_charset == 0)
			e->vc_translate = set_translate(e->vc_G0_charset, vc);
*/
		D("char map G0");
		e->es = ES_NORMAL;
		return;
	case ES_SET_G1:
/*
		if (c == '0')
			e->vc_G1_charset = GRAF_MAP;
		else if (c == 'B')
			e->vc_G1_charset = LAT1_MAP;
		else if (c == 'U')
			e->vc_G1_charset = IBMPC_MAP;
		else if (c == 'K')
			e->vc_G1_charset = USER_MAP;
		if (e->vc_charset == 1)
			e->vc_translate = set_translate(e->vc_G1_charset, vc);
*/
		D("char map G1");
		e->es = ES_NORMAL;
		return;
	default:
		e->es = ES_NORMAL;
	}
}
