#define TERM_NAME "linux"

#ifndef emu_linux_h
#define emu_linux_h

#ifdef HAVE_EMULATION
#error Can not have more than one EMULATION module.
#endif

#define HAVE_EMULATION
#include "emulation/interface.h"

#include BUFFER
#include IO
#include DISPLAY

typedef enum {
	ESnormal,
	ESesc,
	ESsquare,
	ESgetpars,
	ESgotpars,
	ESfunckey,
	EShash,
	ESsetG0,
	ESsetG1,
	ESpercent,
	ESignore,
	ESnonstd,
	ESpalette
} emu_state_t;

typedef enum {
	NUL = 0x00,
	BEL = 0x07,
	BS  = 0x08,
	HT  = 0x09,
	LF  = 0x0a,
	VT  = 0x0b,
	FF  = 0x0c,
	CR  = 0x0d,
	SO  = 0x0e,
	SI  = 0x0f,
	CAN = 0x18,
	SUB = 0x1a,
	ESC = 0x1b,
	DEL = 0x7f,
	CSI = 0x9b,
} emu_control_char_t;

#define NPARAM 16

typedef struct {
	emu_state_t es; // esc seq decode state
	unsigned int nparam;
	unsigned int param[NPARAM]; // esc seq parameters
	emu_mode_t mode; // vt mode
	void *buf; // buffer
	buf_state_t *bs; // cursor, attrs, size
	void *io;
	
	// ouch (scrolling region)
	unsigned int top;
	unsigned int bottom;
	
	// modes
	unsigned int charset; // G0 / G1
	unsigned int s_charset;
	unsigned int disp_ctrl; // display <32
	unsigned int toggle_meta; //toggle high bit
	unsigned int decscnm; // screen
	unsigned int decom;   // origin
	unsigned int decawm;  // autowrap
	unsigned int deccm;   // cursor visible
	unsigned int decim;   // insert
	unsigned int deccolm; // column
	
	unsigned int tab_stop[8];
	// TODO:buf ?
	unsigned char palette[16*3];
	
	unsigned char attr; /* Current attributes */
	unsigned char default_color; /* Default colors */
	unsigned char color; /* Foreground & background */
	unsigned char s_color; /* Saved foreground & background */
	unsigned char ulcolor; /* Color for underline mode */
	unsigned char halfcolor; /* Color for half intensity mode */
	
	//attr
	unsigned int intensity; /* 0=half, 1=normal, 2=bold */
	unsigned int underline;
	unsigned int blink;
	unsigned int reverse;
	
	//saved
	unsigned int s_intensity; /* 0=half, 1=normal, 2=bold */
	unsigned int s_underline;
	unsigned int s_blink;
	unsigned int s_reverse;
	
	int s_x;
	int s_y;
	
	// misc state
	unsigned int ques;
	unsigned int need_wrap;
	
} emu_linux_t;

#endif
